# -*- coding: utf-8 -*-

"""
Core routines and optimization logic.
"""

import collections as _collections
import itertools as _itertools
import json as _json
import logging as _logging

import multiset as _multiset

from . import check as _check


logger = _logging.getLogger(__name__)  # pylint: disable=invalid-name


def convert_order_position(vector):
    """Convert an order vector into a position vector, and vice-versa."""
    # order vectors and position vectors are dual objects, see the definition
    # of final_position for a more detailed description
    dual = [None] * len(vector)
    for i, item in enumerate(vector):
        dual[item] = i
    return dual


def final_position(matrix):
    """Compute the final position of strands according to the linking matrix."""
    assert _check.is_symmetric(matrix)
    size = len(matrix)
    # compute the final order of strands using Melvin's algorithm
    # given a strand `s` and an index `i`, we have `final_order[i] == s`
    final_order = list(range(size))
    for i, row in enumerate(matrix):
        for j in range(i):  # decreasing index in final order
            if row[j] % 2 == 1:
                final_order[i] -= 1
        for j in range(i + 1, size):  # increasing index in final order
            if row[j] % 2 == 1:
                final_order[i] += 1
    # transform final_order to get the final position
    # given a strand `s` and an index `i`, we have `final_position_[s] == i`
    return convert_order_position(final_order)


def _pairwise(iterable):
    # see https://docs.python.org/3/library/itertools.html#itertools-recipes
    a, b = _itertools.tee(iterable)  # pylint: disable=invalid-name
    next(b, None)
    return zip(a, b)


class _OptimizerState:
    """Internal state of the optimization logic to compute crosslevels."""

    __slots__ = ('position', 'crossings', )

    def __init__(self, position, crossings):
        self.position = position  # current position of strands
        self.crossings = crossings  # remaining crossings

    @staticmethod
    def _adjacent_pairs(position):
        """Generate the pairs of adjacent strands, in increasing order."""
        for left, right in _pairwise(position):
            if left < right:
                yield left, right
            else:
                yield right, left

    @staticmethod
    def _is_valid_transition(transition):
        """Return True iff every strand is used in at most one crossing."""
        already_encountered = set()
        for crossing in transition:
            for strand in crossing:
                if strand in already_encountered:
                    return False  # strand is used more than once
                else:
                    already_encountered.add(strand)
        return True

    def _valid_transitions(self):
        """Generate the valid transitions from self."""
        # keep one occurence of each remaining crossings
        crossings_set = set(self.crossings.distinct_elements())
        # only consider crossings of adjacent strands
        crossings_set.intersection_update(self._adjacent_pairs(self.position))

        # yield valid transitions with decreasing number of crossings
        for size in reversed(range(1, len(crossings_set) + 1)):
            yield from filter(
                self._is_valid_transition,
                _itertools.combinations(crossings_set, size)
            )

    @staticmethod
    def _apply_transition(position, transition):
        """Compute the position reached by applying transition to position."""
        order = convert_order_position(position)
        new_position = list(position)  # do not alter original position
        for left, right in transition:
            new_position[order[left]], new_position[order[right]] = \
                    new_position[order[right]], new_position[order[left]]
            order[left], order[right] = order[right], order[left]

        return tuple(new_position)

    def next_states(self):
        """
        Generate the tuples (state, transition) for reachable states from self.
        """
        for transition in self._valid_transitions():
            new_position = self._apply_transition(self.position, transition)
            new_crossings = self.crossings.difference(transition)
            new_state = type(self)(new_position, new_crossings)
            yield new_state, transition

    def __eq__(self, other):
        return self.position == other.position and \
                self.crossings == other.crossings

    def __hash__(self):
        return hash((self.position, self.crossings))


class Template:
    def __init__(self, matrix):
        if not _check.is_linking(matrix):
            raise ValueError('Invalid linking matrix')
        self.matrix = matrix
        self.__crosslevels = None

    @classmethod
    def from_json(cls, fp):  # pylint: disable=invalid-name
        """
        Create a template by loading a linking matrix from a JSON file.

        Deserialize `fp` (a `.read()`-supporting text file containing a JSON
        document) to a Python object compatible with the structure of a linking
        matrix.

        A linking matrix serialized as a JSON document must be structured as a
        JSON array made of arrays of integers.  The validity of the matrix is
        not checked by this function.

        If the data being deserialized has not a structure compatible with the
        structure of a linking matrix, a TypeError will be raised.
        """

        # interpret the input as a JSON file
        try:
            matrix = _json.load(fp)
        except _json.JSONDecodeError as err:
            raise TypeError('Malformed JSON') from err

        # check the loaded input JSON structure is compatible with a linking
        # matrix object: i.e., a List[List[int]]
        if not isinstance(matrix, list):
            raise TypeError('Invalid input structure')
        if not matrix:  # empty linking matrix
            raise TypeError('Invalid input structure')
        for row in matrix:
            if not isinstance(row, list):
                raise TypeError('Invalid input structure')
            for coeff in row:
                if not isinstance(coeff, int):
                    raise TypeError('Invalid input structure')

        return cls(matrix)

    @property
    def size(self):
        """Number of strands of the template."""
        return len(self.matrix)

    @property
    def torsions(self):
        """Number of (oriented) torsions for each strand of the template."""
        return [self.matrix[i][i] for i in range(len(self.matrix))]

    @property
    def crossings(self):
        """
        Mapping of {crossing: arity} for each (oriented) crossing of the template.
        """
        return {
            (j, i): self.matrix[i][j]
            for i in range(len(self.matrix))
            for j in range(i)
            if self.matrix[i][j]  # keep only crossings with non-zero arity
        }

    def _optimize_crosslevels(self):
        """Compute a sequence of crossing levels of minimum depth."""
        initial_state = _OptimizerState(
            tuple(range(self.size)),
            _multiset.FrozenMultiset(
                {
                    # the optimization loop needs the arity of each crossing
                    # without their orientation: drop orientation
                    crossing: abs(arity)
                    for crossing, arity in self.crossings.items()
                }
            )
        )
        final_state = _OptimizerState(
            tuple(final_position(self.matrix)),
            _multiset.FrozenMultiset()
        )

        # depth tracks the number of crosslevels
        # path tracks the previous state
        # how tracks the transition used from the previous state
        depth, path, how = {initial_state: 0}, {}, {}

        lifo = _collections.deque()
        lifo.append(initial_state)
        while lifo:
            current_state = lifo.popleft()
            for new_state, transition in current_state.next_states():
                if new_state not in depth:
                    depth[new_state], path[new_state], how[new_state] = \
                            depth[current_state] + 1, current_state, transition
                    lifo.append(new_state)

        crosslevels = [None] * depth[final_state]
        current_state = final_state
        for depth in reversed(range(depth[final_state])):
            crosslevels[depth] = how[current_state]
            current_state = path[current_state]

        return crosslevels

    @property
    def crosslevels(self):
        """Level-by-level list of concurrent crossings of the template."""
        if self.__crosslevels is None:
            logger.info('Starting optimization of template depth')
            self.__crosslevels = self._optimize_crosslevels()
            logger.info('Finished optimization of template depth')
        return self.__crosslevels



# -*- coding: utf-8 -*-

"""
Logic to handle the Command Line Interface (CLI).
"""

import argparse as _argparse
import json as _json
import logging as _logging
import logging.config as _logging_config
import os as _os
import textwrap as _textwrap

from . import main as _main
from . import __name__ as _pkgname  # name of the main package (i.e., the app)
from . import __version__


logger = _logging.getLogger(__name__)  # pylint: disable=invalid-name


# XXX: SCALE_BOUNDS constant tuple introduces a strong coupling between this
# module and the drawing module: consider defining the constant tuple in the
# drawing module and importing it.
SCALE_BOUNDS = (0.5, 3.0)


def _bounded_float(inf, sup):
    """
    Create a suitable callable for `argparse.add_argument` `type=` argument.

    The returned callable converts the string into a float, and checks the
    converted value falls within [inf, sup] (inclusive set).
    """
    def _type(string):
        value = float(string)
        if inf <= value and value <= sup:
            return value
        else:
            msg = '{} does not fall within [{}, {}].'.format(value, inf, sup)
            raise _argparse.ArgumentTypeError(msg)
    return _type


def _create_parser():
    parser = _argparse.ArgumentParser(
        prog=_pkgname,
        formatter_class=_argparse.RawDescriptionHelpFormatter,
        description='Draw the templates of chaotic attractors.',
        epilog=_textwrap.dedent(
            f'''
            To read a matrix from a file whose name starts with a '-' for example
            '-foo.json', use one of these commands:
              {_pkgname} -- -foo.json

              {_pkgname} ./-foo.json
            '''
        ),
    )
    parser.add_argument(
        '--version',
        action='version',
        version='{} {}'.format(_pkgname, __version__),
    )
    parser.add_argument(
        '-s', '--scale',
        default=1.0,
        type=_bounded_float(*SCALE_BOUNDS),
        help='Alter scale of the template.' \
            ' The scale value must reside between {} and {}.'.format(*SCALE_BOUNDS),
    )
    parser.add_argument(
        '-t', '--complete-flow',
        action='store_true',
        help='Add semicircles depicting the complete flow of the attractor.',
    )
    parser.add_argument(
        '-c', '--no-color',
        action='store_false',
        help='Do not color the template.',
        dest='color',
    )
    parser.add_argument(
        '-o', '--output',
        default='template.svg',
        help='Set the output filename to OUTPUT.' \
            ' Default output filename is \'template.svg\'.' \
            ' Use \'-\' to output the matrix to stdout.',
    )
    parser.add_argument(
        'matrix',
        help='Filename to read the matrix from.' \
            ' The matrix must be encoded as a JSON array of arrays.' \
            ' Use \'-\' to read the matrix from stdin.',
    )
    return parser


def _setup_logging():
    """
    Configure the logging.

    The default logging behavior can be tweaked through the environment
    variable `CATE_LOG_CFG`.
    To modify the configuration of the logging, the environment variable
    `CATE_LOG_CFG` has to point to a JSON file that is a valid configuration
    dictionary.  Do not modify the logging configuration unless you know what
    you are doing!

    See https://docs.python.org/3/library/logging.config.html for more details
    on the configuration of the logging module.
    """
    log_cfg_filename = _os.getenv('CATE_LOG_CFG')
    if log_cfg_filename is not None and _os.path.isfile(log_cfg_filename):
        with open(log_cfg_filename, mode='rt') as log_cfg_fd:
            log_cfg = _json.load(log_cfg_fd)
        _logging_config.dictConfig(log_cfg)
    else:
        _logging.basicConfig(
            # levelname is padded with spaces to the length of the longest levelname
            format='[{levelname:^8}] {message}',
            style='{',
            level=_logging.INFO,
        )


def cli():
    """Entry point for the Command Line Interface (CLI)."""
    _setup_logging()
    parser = _create_parser()
    options = vars(parser.parse_args())  # read argparse.Namespace as a dict
    logger.debug(f'parsed arguments: {options}')
    infile = options.pop('matrix')  # extract positional argument from options
    _main.run(infile, **options)



# -*- coding: utf-8 -*-

import logging as _logging
import sys as _sys

from . import core as _core
from . import export as _export


logger = _logging.getLogger(__name__)  # pylint: disable=invalid-name


def run(infile, *, output, color=True, complete_flow=False, scale=1.0):

    try:
        if infile == '-':  # the special argument '-' means stdin
            template = _core.Template.from_json(_sys.stdin)
        else:
            with open(infile) as fp:  # pylint: disable=invalid-name
                template = _core.Template.from_json(fp)
    except TypeError:
        logger.error("Invalid JSON input")
        exit(code=1)
    except ValueError:
        logger.error("Invalid linking matrix")
        exit(code=1)
    except OSError as err:
        logger.critical(f'Unable to read input: {err}')
        exit(code=2)

    logger.info('Input matrix')
    for row in template.matrix:
        logger.info(f'  {row}')

    logger.info("Starting creation of the SVG template")
    exporter = _export.SVGExporter()
    try:
        if output == '-':  # the special argument '-' means stdout
            exporter.write(
                template,
                output=_sys.stdout,
                color=color,
                complete_flow=complete_flow,
                scale=scale
            )
        else:
            with open(output, mode='w') as fp:  # pylint: disable=invalid-name
                exporter.write(
                    template,
                    output=fp,
                    color=color,
                    complete_flow=complete_flow,
                    scale=scale
                )
    except OSError as err:
        logger.critical(f'Unable to write output: {err}')
        exit(code=2)
    logger.info("Finished creation of the SVG template")



# -*- coding: utf-8 -*-

"""
Template export primitives.
"""

class Exporter:
    _available_exporters = {}

    def __init_subclass__(cls, *, alias=None, **kwargs):
        assert alias not in cls._available_exporters, f'Shadowing existing alias: \'{alias}\''
        super.__init_subclass__(**kwargs)
        if alias is not None:
            cls._available_exporters[alias] = cls

    @classmethod
    def factory(cls, alias):
        try:
            return cls._available_exporters[alias]
        except KeyError:
            raise KeyError(f'Unknown Backend alias: \'{alias}\'') from None

    def write(self, template, *, output, color=True, complete_flow=False, scale=1.0):
        raise NotImplementedError


from ._svg import SVGExporter  # import required for the factory to work



# -*- coding: utf-8 -*-

import io
import pytest
from cate import core
import helpers


# pylint: disable=no-self-use,missing-docstring
class TestFromJSON:
    VALID_INPUTS = {
        'full': '[[1, 1], [1, 2]]',
    }
    MALFORMED_JSON_INPUTS = {
        'empty': '',
        'missing_closing_bracket': '[',
        'extra_comma': '[[0, 0], [0, 0],]',
    }
    INVALID_INPUTS = {
        'None': 'null',
        'boolean': 'true',
        'empty': '[]',
        'dict': '{"super": "matrix"}',
        'string_coeff': '[["1", 0], [0, 1]]',
        'float_coeff': '[[1.0, 0], [0, 1]]',
        'list_of_dict': '[{}]',
    }

    @helpers.dict_parametrize('input_', VALID_INPUTS)
    def test_valid_input(self, input_):
        fp = io.StringIO(input_)  # pylint: disable=invalid-name
        # check no exception is raised
        core.Template.from_json(fp)

    @helpers.dict_parametrize('input_', MALFORMED_JSON_INPUTS)
    def test_malformed_json(self, input_):
        fp = io.StringIO(input_)  # pylint: disable=invalid-name
        with pytest.raises(TypeError, match='^Malformed JSON$'):
            core.Template.from_json(fp)

    @helpers.dict_parametrize('input_', INVALID_INPUTS)
    def test_invalid_input(self, input_):
        fp = io.StringIO(input_)  # pylint: disable=invalid-name
        with pytest.raises(TypeError, match='^Invalid input structure$'):
            core.Template.from_json(fp)



# -*- coding: utf-8 -*-

import itertools
import pytest
from cate import cli
import helpers


@pytest.fixture
def parser():
    return cli._create_parser()


class TestArgumentsParsing:
    def test_missing_matrix(self, parser, capsys):
        with pytest.raises(SystemExit):
            parser.parse_args([])
        stderr = capsys.readouterr().err
        assert 'error: the following arguments are required: matrix' in stderr

    @helpers.dict_parametrize('args', {
        'short': ['-x'],
        'long': ['--wrong-long'],
        'mixed': ['-x', '--wrong-long'],
    })
    def test_unknown_arguments(self, parser, capsys, args):
        with pytest.raises(SystemExit):
            parser.parse_args(args + ['matrix'])
        stderr = capsys.readouterr().err
        assert 'error: unrecognized arguments:' in stderr

    @helpers.dict_parametrize('scale', {
        'non_float': 'ImNoFloat',
        'NaN': 'NaN',
        'infinity': 'Inf',
        'too_small': '0.1',
        'too_big': '10',
    })
    def test_invalid_scale(self, parser, capsys, scale):
        with pytest.raises(SystemExit):
            parser.parse_args(['-s', scale, 'matrix'])
        stderr = capsys.readouterr().err
        assert 'cate: error: argument -s/--scale:' in stderr

    @pytest.mark.parametrize('argname', ['-o', '--output', '-s', '--scale'])
    def test_missing_argvalue(self, parser, capsys, argname):
        with pytest.raises(SystemExit):
            parser.parse_args([argname])
        stderr = capsys.readouterr().err
        assert 'cate: error: argument ' in stderr
        assert ': expected one argument' in stderr

    @pytest.mark.parametrize('combination', itertools.product(
        ('', '-s 1.0', '--scale 1.0'),
        ('', '-t', '--complete-flow'),
        ('', '-c', '--no-color'),
        ('', '-o template.svg', '--output template.svg'),
    ))
    def test_valid_combinations(self, parser, combination):
        args = ' '.join(combination).split()  # needed for arguments with a value
        parser.parse_args(args + ['matrix'])



# -*- coding: utf-8 -*-

import itertools as _itertools

import svgwrite as _svgwrite

from . import core as _core
from . import export as _export


COLORSET = (
    "#1f77b4", "#ff7f0e", "#2ca02c", "#d62728", "#9467bd", "#8c564b",
    "#e377c2", "#7f7f7f",
)


# primitives configuration
STRAND_WIDTH = 40
BOUNDARY_WIDTH = 2
STRAND_GAP = 100  # distance between the left boundary of two neighbor strands
STRETCHING_HEIGHT = 60  # should be > (STRAND_WIDTH - STRAND_GAP) / 2
TORSION_HEIGHT = 40
CROSSING_HEIGHT = 120
LAYERING_HEIGHT = 120
FLOW_GAP = 50


# derived/internal lengths
_MASK_WIDTH = 3 * BOUNDARY_WIDTH
_AA_OVERLAP = 1  # height of layers' overlap to avoid antialiasing artifact
_AA_TORSION_H = TORSION_HEIGHT - _AA_OVERLAP
_AA_CROSSING_H = CROSSING_HEIGHT - _AA_OVERLAP


# boundary curves
_TORSION_LCURVE = \
        f'v {_AA_OVERLAP}' \
        f'c 0 {_AA_TORSION_H / 2} {STRAND_WIDTH} {_AA_TORSION_H / 2} {STRAND_WIDTH} {_AA_TORSION_H}' \
        f'v {_AA_OVERLAP}'
_TORSION_RCURVE = \
        f'v {-_AA_OVERLAP}' \
        f'c 0 {-_AA_TORSION_H / 2} {STRAND_WIDTH} {-_AA_TORSION_H / 2} {STRAND_WIDTH} {-_AA_TORSION_H}' \
        f'v {-_AA_OVERLAP}'
_LSHIFT_CURVE = \
        f'v {_AA_OVERLAP}' \
        f'c 0 {_AA_CROSSING_H / 2} {-STRAND_GAP} {_AA_CROSSING_H / 2} {-STRAND_GAP} {_AA_CROSSING_H}' \
        f'v {_AA_OVERLAP}'
_LSHIFT_REVCURVE = \
        f'v {-_AA_OVERLAP}' \
        f'c 0 {-_AA_CROSSING_H / 2} {STRAND_GAP} {-_AA_CROSSING_H / 2} {STRAND_GAP} {-_AA_CROSSING_H}' \
        f'v {-_AA_OVERLAP}'
_RSHIFT_CURVE = \
        f'v {_AA_OVERLAP}' \
        f'c 0 {_AA_CROSSING_H / 2} {STRAND_GAP} {_AA_CROSSING_H / 2} {STRAND_GAP} {_AA_CROSSING_H}' \
        f'v {_AA_OVERLAP}'
_RSHIFT_REVCURVE = \
        f'v {-_AA_OVERLAP}' \
        f'c 0 {-_AA_CROSSING_H / 2} {-STRAND_GAP} {-_AA_CROSSING_H / 2} {-STRAND_GAP} {-_AA_CROSSING_H}' \
        f'v {-_AA_OVERLAP}'
_LAYERING_CURVE_TEMPLATE = \
        f'c 0 {LAYERING_HEIGHT / 2} {{shift}} {LAYERING_HEIGHT / 2} {{shift}} {LAYERING_HEIGHT}'
_LAYERING_REVCURVE_TEMPLATE = \
        f'c 0 {-LAYERING_HEIGHT / 2} {{shift}} {-LAYERING_HEIGHT / 2} {{shift}} {-LAYERING_HEIGHT}'
_ARC_TEMPLATE = \
        'a 1 1 0 1 0 {width} 0'


class _SVGDrawer:
    def __init__(self):
        self.dwg = _svgwrite.Drawing()
        self.depth = 0
        self._sprites = self._create_sprites()

    def _create_straight_sprite(self, height):
        sprite = self.dwg.g(class_='strand')
        sprite.add(
            self.dwg.rect(
                insert=(0, -_AA_OVERLAP),
                size=(STRAND_WIDTH, height + _AA_OVERLAP),
                class_='interior'
            )
        )
        sprite.add(
            self.dwg.line(
                start=(0, -_AA_OVERLAP),
                end=(0, height),
                class_='boundary'
            )
        )
        sprite.add(
            self.dwg.line(
                start=(STRAND_WIDTH, -_AA_OVERLAP),
                end=(STRAND_WIDTH, height),
                class_='boundary'
            )
        )
        self.dwg.defs.add(sprite)
        return sprite

    def _create_pos_torsion_sprite(self):
        mask = self.dwg.mask()  # emulate the overlap of the boundaries with a mask
        sprite = self.dwg.g(class_='strand')
        mask.add(
            self.dwg.path(
                (f'M 0 {-_AA_OVERLAP}', _TORSION_LCURVE, ),
                stroke='black',
                stroke_width=_MASK_WIDTH,
            )
        )
        mask.add(
            self.dwg.path(
                (f'M 0 {TORSION_HEIGHT}', _TORSION_RCURVE, ),
                stroke='white',
                stroke_width=BOUNDARY_WIDTH,
            )
        )
        sprite.add(
            self.dwg.path(
                (f'M 0 {-_AA_OVERLAP}', _TORSION_LCURVE, f'h {-STRAND_WIDTH}', _TORSION_RCURVE, 'Z', ),
                class_='interior',
            )
        )
        sprite.add(
            self.dwg.path(
                (f'M 0 {-_AA_OVERLAP}', _TORSION_LCURVE, ),
                class_='boundary',
            )
        )
        sprite.add(
            self.dwg.path(
                (f'M 0 {TORSION_HEIGHT}', _TORSION_RCURVE, ),
                mask=mask.get_funciri(),
                class_='boundary',
            )
        )
        self.dwg.defs.add(mask)
        self.dwg.defs.add(sprite)
        return sprite

    def _create_neg_torsion_sprite(self):
        mask = self.dwg.mask()  # emulate the overlap of the boundaries with a mask
        sprite = self.dwg.g(class_='strand')
        mask.add(
            self.dwg.path(
                (f'M 0 {TORSION_HEIGHT}', _TORSION_RCURVE, ),
                stroke='black',
                stroke_width=_MASK_WIDTH,
            )
        )
        mask.add(
            self.dwg.path(
                (f'M 0 {-_AA_OVERLAP}', _TORSION_LCURVE, ),
                stroke='white',
                stroke_width=BOUNDARY_WIDTH,
            )
        )
        sprite.add(
            self.dwg.path(
                (f'M 0 {-_AA_OVERLAP}', _TORSION_LCURVE, f'h {-STRAND_WIDTH}', _TORSION_RCURVE, 'Z', ),
                class_='interior',
            )
        )
        sprite.add(
            self.dwg.path(
                (f'M 0 {-_AA_OVERLAP}', _TORSION_LCURVE, ),
                mask=mask.get_funciri(),
                class_='boundary',
            )
        )
        sprite.add(
            self.dwg.path(
                (f'M 0 {TORSION_HEIGHT}', _TORSION_RCURVE, ),
                class_='boundary',
            )
        )
        self.dwg.defs.add(mask)
        self.dwg.defs.add(sprite)
        return sprite

    def _create_left_shift_sprite(self):
        sprite = self.dwg.g(class_='strand')
        sprite.add(
            self.dwg.path(
                (f'M 0 {-_AA_OVERLAP}', _LSHIFT_CURVE, f'h {STRAND_WIDTH}', _LSHIFT_REVCURVE, 'Z', ),
                class_='interior',
            )
        )
        sprite.add(
            self.dwg.path(
                (f'M 0 {-_AA_OVERLAP}', _LSHIFT_CURVE, ),
                class_='boundary'
            )
        )
        sprite.add(
            self.dwg.path(
                (f'M {STRAND_WIDTH} {-_AA_OVERLAP}', _LSHIFT_CURVE, ),
                class_='boundary'
            )
        )
        self.dwg.defs.add(sprite)
        return sprite

    def _create_right_shift_sprite(self):
        sprite = self.dwg.g(class_='strand')
        sprite.add(
            self.dwg.path(
                (f'M 0 {-_AA_OVERLAP}', _RSHIFT_CURVE, f'h {STRAND_WIDTH}', _RSHIFT_REVCURVE, 'Z', ),
                class_='interior',
            )
        )
        sprite.add(
            self.dwg.path(
                (f'M 0 {-_AA_OVERLAP}', _RSHIFT_CURVE, ),
                class_='boundary'
            )
        )
        sprite.add(
            self.dwg.path(
                (f'M {STRAND_WIDTH} {-_AA_OVERLAP}', _RSHIFT_CURVE, ),
                class_='boundary'
            )
        )
        self.dwg.defs.add(sprite)
        return sprite

    def _create_sprites(self):
        sprites = {
            'no-torsion': self._create_straight_sprite(height=TORSION_HEIGHT),
            'no-shift': self._create_straight_sprite(height=CROSSING_HEIGHT),
            'pos-torsion': self._create_pos_torsion_sprite(),
            'neg-torsion': self._create_neg_torsion_sprite(),
            'left-shift': self._create_left_shift_sprite(),
            'right-shift': self._create_right_shift_sprite(),
        }
        return sprites

    def _use_sprite(self, sprite, strand, position):
        shape = self.dwg.use(
            self._sprites[sprite],
            insert=(position * STRAND_GAP, self.depth),
            class_=f'strand-{strand}',
        )
        return shape

    def set_style(self, palette):
        """
        Set the style of the drawn SVG elements.

        :param palette:
            Assign each strand a color from the iterable palette.  The colors
            are bound to the strands in a round robin fashion.  If palette is
            empty, no colors are bound.
        """
        self.dwg['fill'] = 'white'  # white strands by default
        common_styling = [
            f'.boundary {{fill: none; stroke: black; stroke-width: {BOUNDARY_WIDTH};}}',
        ]
        per_strand_styling = [
            f'.strand-{i} {{fill: {color};}}'
            for i, color in enumerate(palette)
        ]
        style = ' '.join(common_styling + per_strand_styling)
        self.dwg.defs.add(self.dwg.style(style))

    def draw_stretching(self, size):
        """Draw the initial stretching of a template with size strands."""
        shape = self.dwg.g(class_='boundary')
        # draw left border
        shape.add(
            self.dwg.line(
                start=(0, 0),
                end=(0, STRETCHING_HEIGHT),
            )
        )
        # draw right border
        shape.add(
            self.dwg.line(
                start=(STRAND_GAP * (size - 1) + STRAND_WIDTH, 0),
                end=(STRAND_GAP * (size - 1) + STRAND_WIDTH, STRETCHING_HEIGHT),
            )
        )
        # draw stretching of each strand
        for position in range(size - 1):
            shape.add(
                self.dwg.path(
                    (
                        f'M {STRAND_GAP * (position + 1)} {STRETCHING_HEIGHT}',
                        f'v {-_AA_OVERLAP}',
                        _ARC_TEMPLATE.format(width=STRAND_WIDTH - STRAND_GAP),
                        f'v {_AA_OVERLAP}',
                    ),
                )
            )
        self.dwg.add(shape)

    def draw_torsion(self, strand, position, torque):
        """Draw a single torsion for a given strand."""
        if torque == 0:
            sprite = 'no-torsion'
        elif torque > 0:
            sprite = 'pos-torsion'
        else:  # torque < 0
            sprite = 'neg-torsion'

        shape = self._use_sprite(sprite, strand, position)
        self.dwg.add(shape)

    def draw_no_crossing(self, strand, position):
        """Draw a straight transition the height of a crossing."""
        shape = self._use_sprite('no-shift', strand, position)
        self.dwg.add(shape)

    def draw_crossing(self, left, right, orientation):
        """Draw the crossing of two strands."""
        lstrand, lpos = left
        rstrand, rpos = right
        if lpos > rpos:
            # ensure lstrand, lpos point to the leftmost strand
            lstrand, rstrand = rstrand, lstrand
            lpos, rpos = rpos, lpos

        shape = self.dwg.g()
        lshift = self._use_sprite('left-shift', rstrand, rpos)
        rshift = self._use_sprite('right-shift', lstrand, lpos)
        if orientation:
            # a positive crossing shows the right strand (left shift) above
            shape.add(rshift)
            shape.add(lshift)
        else:
            # a negative crossing shows the left strand (right shift) above
            shape.add(lshift)
            shape.add(rshift)
        self.dwg.add(shape)

    def draw_layer(self, strand, position, size):
        """Draw the layering step for a given strand."""
        lshift = position * STRAND_GAP
        rshift = (size - 1 - position) * STRAND_GAP
        shape = self.dwg.g(class_=f'strand strand-{strand}')
        shape.add(
            self.dwg.path(
                (
                    f'M {lshift} {self.depth - _AA_OVERLAP}',
                    f'v {_AA_OVERLAP}',
                    _LAYERING_CURVE_TEMPLATE.format(shift=-lshift),
                    f'h {(size - 1) * STRAND_GAP + STRAND_WIDTH}',
                    _LAYERING_REVCURVE_TEMPLATE.format(shift=-rshift),
                    f'v {-_AA_OVERLAP}',
                    'Z',
                ),
                class_='interior',
            )
        )
        shape.add(
            self.dwg.path(
                (
                    f'M {lshift} {self.depth - _AA_OVERLAP}',
                    f'v {_AA_OVERLAP}',
                    _LAYERING_CURVE_TEMPLATE.format(shift=-lshift),
                ),
                class_='boundary',
            )
        )
        shape.add(
            self.dwg.path(
                (
                    f'M {lshift + STRAND_WIDTH} {self.depth - _AA_OVERLAP}',
                    f'v {_AA_OVERLAP}',
                    _LAYERING_CURVE_TEMPLATE.format(shift=rshift),
                ),
                class_='boundary',
            )
        )
        self.dwg.add(shape)

    def draw_flow(self, size, style='minimal'):
        """Draw the template flow."""
        if style == 'minimal':
            shape = self.dwg.g(class_='boundary', stroke_linecap='square')
            shape.add(
                self.dwg.line(
                    start=(0, 0),
                    end=(STRAND_GAP * (size - 1) + STRAND_WIDTH, 0),
                )
            )
            shape.add(
                self.dwg.line(
                    start=(0, self.depth),
                    end=(STRAND_GAP * (size - 1) + STRAND_WIDTH, self.depth),
                )
            )
        elif style == 'complete':
            shape = self.dwg.g(class_='boundary')
            shape.add(
                self.dwg.path(
                    (
                        'M 0 0',
                        _ARC_TEMPLATE.format(width=-FLOW_GAP),
                        f'v {self.depth}',
                        _ARC_TEMPLATE.format(width=FLOW_GAP),
                    ),
                )
            )
            width = STRAND_GAP * (size - 1) + STRAND_WIDTH
            shape.add(
                self.dwg.path(
                    (
                        f'M {width} 0',
                        _ARC_TEMPLATE.format(width=-FLOW_GAP - 2 * width),
                        f'v {self.depth}',
                        _ARC_TEMPLATE.format(width=FLOW_GAP + 2 * width),
                    ),
                )
            )
        else:
            raise ValueError(f'Unknown style \'{style}\'')
        self.dwg.add(shape)


class SVGExporter(_export.Exporter, alias='svg'):
    def __init__(self, colorset=COLORSET):
        self.positions = None
        self.drawer = None
        self.colorset = colorset

    def _palette(self, size, color=True):
        # pylint: disable=no-else-return
        if color:
            # cycle through COLORS as long as needed
            return _itertools.islice(_itertools.cycle(self.colorset), size)
        else:
            return ()

    def _export_stretching(self, size):
        self.drawer.draw_stretching(size)
        self.drawer.depth += STRETCHING_HEIGHT

    @staticmethod
    def _decrease_torsion_torque(torque):
        # pylint: disable=no-else-return
        if torque == 0:
            return 0
        elif torque < 0:
            return torque + 1
        else:
            return torque - 1

    def _export_torsions(self, template):
        torsions = list(template.torsions)  # work on a copy
        while any(torsions):  # at least one strand remains twisted
            for strand, torque in enumerate(torsions):
                self.drawer.draw_torsion(strand, strand, torque)
                torsions[strand] = self._decrease_torsion_torque(torque)
            self.drawer.depth += TORSION_HEIGHT

    def _export_crossings(self, template):
        for crosslevel in template.crosslevels:
            uncrossed = set(range(template.size))  # track uncrossed strands
            for crossing in crosslevel:
                lstrand, rstrand = crossing
                lpos, rpos = self.positions[lstrand], self.positions[rstrand]
                self.drawer.draw_crossing(
                    left=(lstrand, lpos),
                    right=(rstrand, rpos),
                    orientation=template.crossings[lstrand, rstrand] > 0
                )
                uncrossed -= {lstrand, rstrand}
                self.positions[lstrand], self.positions[rstrand] = \
                    self.positions[rstrand], self.positions[lstrand]
            for strand in uncrossed:  # draw remaining uncrossing strands
                self.drawer.draw_no_crossing(strand, self.positions[strand])
            self.drawer.depth += CROSSING_HEIGHT

    def _export_layering(self, size):
        # paint layers from left to right (according to order, not position)
        final_order = _core.convert_order_position(self.positions)
        for position, strand in enumerate(final_order):
            self.drawer.draw_layer(strand, position, size)
        self.drawer.depth += LAYERING_HEIGHT

    def _export_flow(self, size, complete_flow=False):
        style = 'complete' if complete_flow else 'minimal'
        self.drawer.draw_flow(size, style=style)

    def _export_initialize(self, template, color=True):
        size = template.size
        self.drawer = _SVGDrawer()
        self.positions = list(range(size))
        self.drawer.set_style(self._palette(size, color=color))
        self._export_stretching(size)

    def _export_finalize(self, template, complete_flow=False):
        size = template.size
        self._export_layering(size)
        self._export_flow(size, complete_flow=complete_flow)

    def write(self, template, *, output, color=True, complete_flow=False, scale=1.0):
        self._export_initialize(template, color=color)
        self._export_torsions(template)
        self._export_crossings(template)
        self._export_finalize(template, complete_flow=complete_flow)
        self.drawer.dwg.write(output)



# -*- coding: utf-8 -*-

"""
CATE: Chaotic Attractor TEmplate


Chaotic attractors are solutions of deterministic processes, of which the
topology can be described by templates.  We consider templates of chaotic
attractors bounded by a genus-1 torus described by a linking matrix.

This package provides modules to:
    - validate a linking matrix by checking continuity and determinism
      constraints.
    - draw a template of a valid linking matrix.  The representation of the
      template is compact, as the height of the template is minimized.  The
      representation is saved as a Scalable Vector Graphics (SVG) file.

This package also provides a CLI.
"""

__version__ = '1.0b0'



# -*- coding: utf-8 -*-

import pytest


def dict_parametrize(argnames, paramsdict, indirect=False, scope=None):
    """Decorator to parametrize test functions from a (id, argvalue) dict."""
    ids, argvalues = zip(*paramsdict.items())  # ensure id matches its argvalue
    return pytest.mark.parametrize(argnames, argvalues, indirect, ids, scope)



# -*- coding: utf-8 -*-

import pytest
from cate import check


# pylint: disable=missing-docstring
class TestCheck:
    TEST_MATRICES = {
        '2x2_fail_square': ((0, 0), (0, )),
        '2x3_fail_square': ((0, 1, 2), (3, 4, 5)),
        '3x3_fail_symmetric': ((-1, 0, -1), (-1, 0, 0), (-1, 0, 1)),
        '2x2_fail_diagonal': ((0, 0), (0, 2)),
        '4x4_fail_neighbors': ((3, 2, 2, 3), (2, 2, 2, 3), (2, 2, 3, 4), (3, 3, 4, 4)),
        '4x4_fail_final-position': ((0, 0, 0, -1), (0, 1, 0, 0), (0, 0, 0, -1), (-1, 0, -1, -1)),
        '2x2-0_fail_submatrix': ((0, 1), (1, 1)),
        '2x2-1_fail_submatrix': ((1, 2), (2, 2)),
        '2x2-2_fail_submatrix': ((-1, -1), (-1, -2)),
        '3x3_fail_submatrix': ((-1, -1, -1), (-1, -2, -2), (-1, -2, -3)),
        '4x4_fail_planarity': ((0, 0, 0, 0), (0, 1, 0, -1), (0, 0, 0, -1), (0, -1, -1, -1)),
        '2x2_ok': ((0, 0), (0, 1)),
        '4x4_ok': ((-1, -1, -1, -1), (-1, 0, 0, 0), (-1, 0, 1, 1), (-1, 0, 1, 2)),
    }

    @pytest.mark.parametrize('input_', (
        '3x3_fail_symmetric',
        '2x2_fail_diagonal',
        '4x4_fail_neighbors',
        '4x4_fail_final-position',
        '2x2-0_fail_submatrix',
        '2x2-1_fail_submatrix',
        '2x2-2_fail_submatrix',
        '3x3_fail_submatrix',
        '4x4_fail_planarity',
        '2x2_ok',
        '4x4_ok',
    ))
    def test_valid_is_square(self, input_):
        assert check.is_square(self.TEST_MATRICES[input_])

    @pytest.mark.parametrize('input_', (
        '2x2_fail_square',
        '2x3_fail_square',
    ))
    def test_ko_is_square(self, input_):
        assert not check.is_square(self.TEST_MATRICES[input_])

    @pytest.mark.parametrize('input_', (
        '2x2_fail_diagonal',
        '4x4_fail_neighbors',
        '4x4_fail_final-position',
        '2x2-0_fail_submatrix',
        '2x2-1_fail_submatrix',
        '2x2-2_fail_submatrix',
        '3x3_fail_submatrix',
        '4x4_fail_planarity',
        '2x2_ok',
        '4x4_ok',
    ))
    def test_valid_is_symmetric(self, input_):
        assert check.is_symmetric(self.TEST_MATRICES[input_])

    @pytest.mark.parametrize('input_', (
        '3x3_fail_symmetric',
    ))
    def test_ko_is_symmetric(self, input_):
        assert not check.is_symmetric(self.TEST_MATRICES[input_])

    @pytest.mark.parametrize('input_', (
        '3x3_fail_symmetric',
        '4x4_fail_neighbors',
        '4x4_fail_final-position',
        '2x2-0_fail_submatrix',
        '2x2-1_fail_submatrix',
        '2x2-2_fail_submatrix',
        '3x3_fail_submatrix',
        '4x4_fail_planarity',
        '2x2_ok',
        '4x4_ok',
    ))
    def test_valid_diagonal_criterion(self, input_):
        assert check._diagonal_criterion(self.TEST_MATRICES[input_])  # pylint: disable=protected-access

    @pytest.mark.parametrize('input_', (
        '2x2_fail_diagonal',
    ))
    def test_ko_diagonal_criterion(self, input_):
        assert not check._diagonal_criterion(self.TEST_MATRICES[input_])  # pylint: disable=protected-access

    @pytest.mark.parametrize('input_', (
        '4x4_fail_final-position',
        '2x2-0_fail_submatrix',
        '2x2-1_fail_submatrix',
        '2x2-2_fail_submatrix',
        '3x3_fail_submatrix',
        '4x4_fail_planarity',
        '2x2_ok',
        '4x4_ok',
    ))
    def test_valid_neighbors_criterion(self, input_):
        assert check._neighbors_criterion(self.TEST_MATRICES[input_])  # pylint: disable=protected-access

    @pytest.mark.parametrize('input_', (
        '2x2_fail_diagonal',
        '4x4_fail_neighbors',
    ))
    def test_ko_neighbors_criterion(self, input_):
        assert not check._neighbors_criterion(self.TEST_MATRICES[input_])  # pylint: disable=protected-access

    @pytest.mark.parametrize('input_', (
        '2x2_fail_diagonal',
        '2x2-0_fail_submatrix',
        '2x2-1_fail_submatrix',
        '2x2-2_fail_submatrix',
        '3x3_fail_submatrix',
        '4x4_fail_planarity',
        '2x2_ok',
        '4x4_ok',
    ))
    def test_valid_final_position_criterion(self, input_):
        assert check._final_position_criterion(self.TEST_MATRICES[input_])  # pylint: disable=protected-access

    @pytest.mark.parametrize('input_', (
        '4x4_fail_neighbors',
        '4x4_fail_final-position',
    ))
    def test_ko_final_position_criterion(self, input_):
        assert not check._final_position_criterion(self.TEST_MATRICES[input_])  # pylint: disable=protected-access

    @pytest.mark.parametrize('input_', (
        '2x2-0_fail_submatrix',
        '2x2-1_fail_submatrix',
        '2x2-2_fail_submatrix',
        '3x3_fail_submatrix',
        '4x4_fail_planarity',
        '2x2_ok',
        '4x4_ok',
    ))
    def test_valid_is_continuous(self, input_):
        assert check.is_continuous(self.TEST_MATRICES[input_])

    @pytest.mark.parametrize('input_', (
        '2x2_fail_diagonal',
        '4x4_fail_neighbors',
        '4x4_fail_final-position',
    ))
    def test_ko_is_continuous(self, input_):
        assert not check.is_continuous(self.TEST_MATRICES[input_])

    @pytest.mark.parametrize('input_', (
        '2x3_fail_square',
        '3x3_fail_symmetric',
        '2x2_fail_diagonal',
        '4x4_fail_final-position',
        '4x4_fail_planarity',
        '2x2_ok',
        '4x4_ok',
    ))
    def test_valid_submatrix_criterion(self, input_):
        assert check._submatrix_criterion(self.TEST_MATRICES[input_])  # pylint: disable=protected-access

    @pytest.mark.parametrize('input_', (
        '4x4_fail_neighbors',
        '2x2-0_fail_submatrix',
        '2x2-1_fail_submatrix',
        '2x2-2_fail_submatrix',
        '3x3_fail_submatrix',
    ))
    def test_ko_submatrix_criterion(self, input_):
        assert not check._submatrix_criterion(self.TEST_MATRICES[input_])  # pylint: disable=protected-access

    @pytest.mark.parametrize('input_', (
        '2x2_fail_diagonal',
        '2x2_ok',
        '4x4_ok',
    ))
    def test_valid_planarity_criterion(self, input_):
        assert check._planarity_criterion(self.TEST_MATRICES[input_])  # pylint: disable=protected-access

    @pytest.mark.parametrize('input_', (
        '4x4_fail_planarity',
    ))
    def test_ko_planarity_criterion(self, input_):
        assert not check._planarity_criterion(self.TEST_MATRICES[input_])  # pylint: disable=protected-access

    @pytest.mark.parametrize('input_', (
        '2x2_ok',
        '4x4_ok',
    ))
    def test_valid_is_deterministic(self, input_):
        assert check.is_deterministic(self.TEST_MATRICES[input_])

    @pytest.mark.parametrize('input_', (
        '2x2-0_fail_submatrix',
        '2x2-1_fail_submatrix',
        '2x2-2_fail_submatrix',
        '3x3_fail_submatrix',
        '4x4_fail_planarity',
    ))
    def test_ko_is_deterministic(self, input_):
        assert not check.is_deterministic(self.TEST_MATRICES[input_])

    @pytest.mark.parametrize('input_', (
        '2x2_ok',
        '4x4_ok',
    ))
    def test_valid_is_linking(self, input_):
        assert check.is_linking(self.TEST_MATRICES[input_])

    @pytest.mark.parametrize('input_', (
        '2x3_fail_square',
        '3x3_fail_symmetric',
        '2x2_fail_diagonal',
        '4x4_fail_neighbors',
        '4x4_fail_final-position',
        '2x2-0_fail_submatrix',
        '2x2-1_fail_submatrix',
        '2x2-2_fail_submatrix',
        '3x3_fail_submatrix',
        '4x4_fail_planarity',
    ))
    def test_ko_is_linking(self, input_):
        assert not check.is_linking(self.TEST_MATRICES[input_])



# -*- coding: utf-8 -*-

"""
Logic to check the validity of linking matrices.
"""

import itertools as _itertools

from . import core as _core


def is_square(matrix):
    """Check whether matrix is square."""
    size = len(matrix)
    return all(len(row) == size for row in matrix)


def is_symmetric(matrix):
    """Check whether matrix is symmetric."""
    assert is_square(matrix)
    size = len(matrix)
    transpose_pairs = (
        (matrix[i][j], matrix[j][i])
        for i in range(size)
        for j in range(i)
    )
    return all(x == y for x, y in transpose_pairs)


def _diagonal_criterion(matrix):
    """Check whether neighbor elements on the diagonal differ by exactly one."""
    size = len(matrix)
    diagonal_pairs = (
        (matrix[i][i], matrix[i + 1][i + 1])
        for i in range(size - 1)
    )
    return all(x - y in (-1, 1) for x, y in diagonal_pairs)


def _neighbors_criterion(matrix):
    """Check whether neighbor elements in the matrix differ by at most one."""
    assert is_symmetric(matrix)
    size = len(matrix)
    orthogonal_pairs = (
        # checking verticals for the whole matrix is enough due to symmetry
        (matrix[i][j], matrix[i + 1][j])
        for i in range(size - 1)
        for j in range(size)
    )
    diagonal_pairs = (
        (matrix[i][j], matrix[i + 1][j + 1])
        # checking lower triangular matrix is enough due to symmetry
        for i in range(1, size - 1)  # ignore diagonal as it as already been checked
        for j in range(i)
    )
    antidiagonal_pairs = (
        (matrix[i][j], matrix[i - 1][j + 1])
        # checking lower triangular matrix is enough due to symmetry
        for i in range(2, size)
        for j in range(i - 1)
    )
    neighbor_pairs = _itertools.chain(
        orthogonal_pairs,
        diagonal_pairs,
        antidiagonal_pairs
    )
    return all(x - y in (-1, 0, 1) for x, y in neighbor_pairs)


def _final_position_criterion(matrix):
    """Check whether the final position of strands is a genuine permutation."""
    size = len(matrix)
    return set(_core.final_position(matrix)) == set(range(size))


def is_continuous(matrix):
    """Check whether the linking matrix is continuous, i.e. no strands are torn."""
    return (
        _diagonal_criterion(matrix)
        and _neighbors_criterion(matrix)
        and _final_position_criterion(matrix)
    )


def _submatrix_criterion(matrix):
    """Check adjacent strands overlap deterministically."""
    size = len(matrix)
    # A 2x2 submatrix | x y | is encoded as the tuple (x, y, z, t).
    #                 | z t |
    diagonal_submatrices = (
        (matrix[i][i], matrix[i][i + 1], matrix[i + 1][i], matrix[i + 1][i + 1])
        for i in range(size - 1)
    )
    # The presence of submatrices | -1 0 | or | 0  0 | (up to a constant
    #                             |  0 0 |    | 0 -1 |
    # additive factor) indicates a violation of determinism.  For a 2x2
    # submatrix encoded as the tuple (x, y, z, t), this is the case when:
    #     x + 1 == y == z == t or x == y == z == t + 1
    return not any(
        x + 1 == y == z == t or x == y == z == t + 1
        for x, y, z, t in diagonal_submatrices
    )


def _planarity_criterion(matrix):
    """Check whether the final position of strands is planar."""

    # We associate to each strand i two intervals: [i - 1, i] and [i, i + 1].
    # We only consider for strand 0 (resp. size - 1) the interval [0, 1]
    # (resp.  [size - 2, size - 1]).  Each interval is identified by its lower
    # (left) bound.
    #
    # The final position of the strands represents a sequence of intervals
    # opening/closing.  The first occurrence of interval i is an opening, the
    # second is a closing.
    #
    # Checking whether the final position of strands is planar is then
    # equivalent to check whether both the sequence of intervals with even
    # identifiers and the sequence of intervals with odd identifiers correspond
    # to a valid nesting of function calls (opening is a function call, closing
    # is a function return).  This is done with a sweep-line algorithm.

    def _intervals_sequence(final_position_, *, parity):
        """Generate the sequence of intervals with a given parity."""
        assert parity in (0, 1)
        for bound in final_position_:
            if bound % 2 == parity:
                if bound < len(final_position_) - 1:
                    yield bound  # open interval
            else:
                if bound > 0:
                    yield bound - 1  # close interval

    def _is_valid_intervals_sequence(sequence):
        """Check whether sequence represents a valid nesting of function calls."""
        history = set()  # set of already encountered intervals
        stack = []  # stack of active intervals (opened but no closed)
        for interval in sequence:
            if interval in history:
                try:
                    active_interval = stack.pop()
                except IndexError:
                    return False  # an already closed interval has been re-opened
                if interval != active_interval:
                    return False
            else:
                history.add(interval)
                stack.append(interval)
        return not bool(stack)  # all intervals must be closed

    final_position_ = _core.final_position(matrix)
    even_intervals = _intervals_sequence(final_position_, parity=0)
    odd_intervals = _intervals_sequence(final_position_, parity=1)
    return (
        _is_valid_intervals_sequence(even_intervals)
        and _is_valid_intervals_sequence(odd_intervals)
    )


def is_deterministic(matrix):
    """Check whether the linking matrix is deterministic."""
    return (
        _submatrix_criterion(matrix)
        and _planarity_criterion(matrix)
    )


def is_linking(matrix):
    """Check whether the matrix is a valid linking matrix."""
    return (
        is_square(matrix)
        and is_symmetric(matrix)
        and is_continuous(matrix)
        and is_deterministic(matrix)
    )
